"
    Copyright (c) 2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"
"{ Package: 'jv:dragonfly' }"

"{ NameSpace: Smalltalk }"

TestCase subclass:#CompiledCodeObjectExamples
	instanceVariableNames:''
	classVariableNames:''
	poolDictionaries:''
	category:'System-Compiler-Interface-Tests'
!

!CompiledCodeObjectExamples class methodsFor:'documentation'!

copyright
"
    Copyright (c) 2016-now Jan Vrany <jan.vrany [at] fit . cvut . cz>

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the 'Software'), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

"
! !

!CompiledCodeObjectExamples class methodsFor:'testing'!

isTestSelector:aSelector
    ^ (super isTestSelector:aSelector) or:[ aSelector startsWith: 'example' ]

    "Created: / 16-06-2016 / 00:23:02 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!CompiledCodeObjectExamples methodsFor:'examples'!

example01
    "Shows how to:
      * allocate a machine code object, 
      * install some code and
      * run it"

    | method object machineCode |        

    "We need to create a method for which we allocate a code object.
     Even though we may not install the method anywhere, we still need it
     as the Method instance keeps the reference to the code object and
     therefore prevents it from being garbage collected."
    method := Method new.
    method numberOfArgs: 0.
    method numberOfVars: 0.
    method stackSize: 0. 

    "Now allocate a code object for the method. We don't know the size of the code yet
     (so we pass 0 as .text section size) and - in this example - we won't need any ILC
     nor literal vector. Remember we can allocate the .text section later once we know
     the size of the machine code."
    object := CompiledCodeObject forCompiledCode: method text: 0 literals: 0 ilcs: 0. 

    "Generate some machine code:"
    machineCode := #[184 11 0 0 0 195].
    "This was easy, wasn't it? The above is
        mov eax, 0xb
        ret
     i.e., a super-simple function that returns a SmallInteger value of 5.    

     Now we need to allocate a .text section and copy the machine code there:"
    object allocateTextSection: machineCode size.           
    object text replaceFrom: 1 with: machineCode.

    "Almost there. The final step is to update a code pointer of the method to
     the machine code we just installed:"
    method code: object text address.

    "Let's execute it:"
    self assert: 5 == (method valueWithReceiver:1 arguments:#()).

    "Created: / 16-06-2016 / 00:23:25 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (format): / 16-06-2016 / 09:29:45 / Jan Vrany <jan.vrany@fit.cvut.cz>"
!

example02
    "Shows how to access a constant object (literal) within a machine code.
     As object may be moved around by a garbage collector, we cannot simply
     take object's address and return it - it will soon become invalid.

     So, to access a constant object from a machine code, we have to allocate
     a special 'object vector' section and put the object there and access it
     indirectly using object's index in the vector and section base address.
     By marking the section as 'object vector' section, the garbage collector
     knows it contains objects and whenever it moves an object referenced from
     there, it updates the pointer in the section data. Note, that section
     data never moves.
    "

    | literal method object asm machineCode |        

    "Create a constant object to return from machine code"
    literal := 'Some cool literal' copy.

    method := Method new.
    method numberOfArgs: 0.
    method numberOfVars: 0.
    method stackSize: 0. 

    "Now allocate a code object for the method. We don't know the size of the code yet
     (so we pass 0 as .text section size). We need to reference our `literal` object,
     so allocate 'object vector' section with size 1 (we need only one object). Still
     no ILCs."
    object := CompiledCodeObject forCompiledCode: method text: 0 literals: 1 ilcs: 0.

    "Store our literal object into the object vector section"
    object literals at: 1 put: literal.

    "Actually, generate the machine code. Do it with help of Igor Stasenko AsmJIT, 
     it iss easier to see it in symbolic instructions rather than machine code"
    self skipIf: (OperatingSystem getCPUType ~~ #x86) description: 'AsmJIT works fine only for i386 assembly :-('.
    self skipIf: (Smalltalk includesKey: #AJx86Assembler) not description: 'AsmJIT not available'.

    asm := AJx86Assembler noStackFrame.
    asm mov: (object literals address asUImm32 ptr) -> asm EAX.
    asm ret.
    machineCode := asm bytes.
    "
    To check the generated code, inspect
        UDIS86 disassemble: machineCode
    "

    "Now we need to allocate a .text section and copy the machine code there:"
    object allocateTextSection: machineCode size.           
    object text replaceFrom: 1 with: machineCode.

    "Almost there. The final step is to update a code pointer of the method to
     the machine code we just installed:"
    method code: object text address.

    "Let's execute it:"
    self assert: literal == (method valueWithReceiver:1 arguments:#()).

    "Created: / 16-06-2016 / 12:29:53 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 16-06-2016 / 17:22:37 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!CompiledCodeObjectExamples class methodsFor:'documentation'!

version_HG

    ^ '$Changeset: <not expanded> $'
! !

